
module Compilers.Vim where

import qualified Data.Set as Set
import qualified Data.List as List

import Syntax.Position
import Syntax.Internal
import Utils.Files
import Utils.Unicode
import Utils.Pairs
import Utils.Lists
import Utils.Functions

vimFile :: FilePath -> FilePath
vimFile file =
    case splitFilePath file of
	(path,name,ext)	-> path ++ "." ++ name ++ ext ++ ".vim"

escape = concatMap esc . nameString
    where
	escchars = "$\\^.*~"
	esc c	| c `elem` escchars = ['\\',c]
		| otherwise	    = [c]

keyword _ [] = ""
keyword cat ws	= "syn keyword " ++ unwords (cat : map nameString ws)
match _ [] = ""
match cat ws	= "syn match " ++ cat ++ " \"" ++
		    concat (List.intersperse "\\|" $ map escape ws) ++ "\""

matches typs cons axs defs =
    map snd
    $ List.sortBy (compare **.* fst)
    $ typs' ++ cons' ++ axs' ++ defs'
    where
	typs' = foo "agdaInfixType" typs
	cons' = foo "agdaInfixConstructor" cons
	axs'  = foo "agdaInfixAxiom" axs
	defs' = foo "agdaInfixFunction" defs
	
	foo cat = map (length . nameString . head /\ match cat)

toVim :: Names -> String
toVim names =
    unlines $
    [ keyword "agdaType"	typs
    , keyword "agdaConstructor" cons
    , keyword "agdaAxiom"	axs
    , keyword "agdaFunction"    defs
    ] ++ matches (classify len ityps)
		 (classify len icons)
		 (classify len iaxs)
		 (classify len idefs)
    where
	len = length . nameString
	ns = Set.toList names

	isQuote = isQuote' . nameString
	    where
		isQuote' ('"':_) = True
		isQuote' _	 = False

	typs  = [ x | ITyp _ x <- ns, not $ isOperator x, not $ isQuote x ]
	cons  = [ x | ICon _ x <- ns, not $ isOperator x, not $ isQuote x ]
	defs  = [ x | IDef _ x <- ns, not $ isOperator x, not $ isQuote x ]
	axs   = [ x | IAx  _ x <- ns, not $ isOperator x, not $ isQuote x ]
	ityps = [ x | ITyp _ x <- ns, isOperator x, not $ isQuote x ]
	icons = [ x | ICon _ x <- ns, isOperator x, not $ isQuote x ]
	idefs = [ x | IDef _ x <- ns, isOperator x, not $ isQuote x ]
	iaxs  = [ x | IAx  _ x <- ns, isOperator x, not $ isQuote x ]

type Names = Set.Set DefName

data DefName = IDef { range :: Range, getName :: Name }
	     | IAx  { range :: Range, getName :: Name }
	     | ICon { range :: Range, getName :: Name }
	     | ITyp { range :: Range, getName :: Name }

instance Eq DefName where
    x == y  = getName x == getName y

instance Ord DefName where
    compare x y = compare (getName x) (getName y)

definedNames :: Program -> [DefName]
definedNames = concatMap defName
    where
	defName (NormalDef r x _ _ _ cs)    = IDef r x : concatMap clauseNames cs
	defName (NewAxiom r x _ _)	    = [IAx r x]
	defName (NewData _ r x _ _ _ cs)    = ITyp r x : map cons cs
	defName (NewSig r x _ _ _ _ cs)	    = [ITyp r x]
	defName (Rewrite _ _)		    = []
	defName (Mutual _ ds)		    = concatMap defName ds
	defName (AbstractDefs _ ds)	    = concatMap defName ds
	defName (Synonym r x _)		    = [IDef r x]

	clauseNames (Clause _ _ _ with) = withNames with

	withNames (NoWith _)	= []
	withNames AbsurdClause	= []
	withNames (With x _ cs) = noDontCare x ++ concatMap clauseNames cs
	    where
		noDontCare DontCare = []
		noDontCare x	    = [IDef noRange x]

	cons (Constructor r x _ _)  = ICon r x

generateVimFile :: FilePath -> Program -> IO ()
generateVimFile file p = writeFileUTF8 (vimFile file) $ toVim names
    where
	names = foldr Set.insert Set.empty $ definedNames p

